/***************************************************************************************************
*
*	$Header::                                                                                      $
*
*	PDBPlus.plw	-	A replacement PDB loader for IDA Pro 4.5 that uses newer versions of DBGHELP.DLL
*					to get all local variable names into an IDA database, in addition to standard
*					function/data names. In future this might be extended to define IDA structures
*					and enumerations based on the data in the .PDB file.. but I need to find time
*					first.
*
*	NOTES			This was initially based on the PDB plugin source code in the IDA Pro 4.5 SDK,
*					which was by Ilfak Guilfanov, and which had changes by Christopher Li. As 
*					opposed to the original version, this *requires* a recent version of DBGHELP.DLL
*					as I didn't like working with code that had to support antiquated versions of
*					the libraries. So, while I was formatting the source to suit my own style I 
*					scrapped the stuff that I didn't need.
*
*					This source is formatted to for on a >100 column wide display, with tabs of 4.
*					The project supplied is for Visual Studio .NET..					
*
*	CHANGES
*
*	23/2/2003	Dean	Created
*
***************************************************************************************************/

#include "stdafx.h"
#include <dbghelp.h>

#pragma warning( disable : 4800 )	// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4099 )	// IDA references an internal type as both class and struct.. 
									// I don't like warnings, so I've pragma'd it away for now.	
#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <auto.hpp>

// We define this so we don't need to bundle Microsoft's 'cvconst.h' file. 
#define	SymTagFunction 5

// Global data
int		giGlobalFunctionCount;
int		giGlobalVariableCount;
int		giLocalVariableCount;
void*	ghProcess = (void *)0xBEEFF00D;
int		giDelta;

enum	ENUMSYM_MODE
{
	ENUMERATE_GLOBALS,
	ENUMERATE_LOCALS,
};

ENUMSYM_MODE	geEnumerationMode;


/***************************************************************************************************
*
*	FUNCTION		EnumerateLocalSymbolsProc
*
*	DESCRIPTION		This callback is used to process symbols that exist within the scope of a 
*					particular function. In our case we're using it to look for register relative
*					symbols, as this gets us local variables that are stored as offsets from BP.
*
*	INPUTS			pSymInfo		-		A filled in SYMBOL_INFO structure.
*	
*					ulSymbolSize	-		Size of the symbol we're processing.
*
*					pvEA			-		This is the delta-adjusted effective address of the
*											function we're processing local symbols for.
*
*	RESULT			This function should return 'FALSE' if, for any reason, we need to stop the
*					symbol enumeration. Otherwise return 'TRUE'.
*
*	NOTES			I've encountered local symbols that are not marked as register relative. These
*					typically contain real addresses in the pSymInfo->Address field (rather than 
*					a frame offset). I believe these are static locals, which exist within the .data
*					or .rdata segments.. in future I might see if I can create names for those
*					variables too.
*
***************************************************************************************************/

static BOOL CALLBACK EnumerateLocalSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvEA )
{
	ea_t	ea		= ( ea_t )( pvEA );
	int		iOffset	= ( int )( ( DWORD )pSymInfo->Address );

	// We only resolve register relative offsets... 
	if ( pSymInfo->Flags & IMAGEHLP_SYMBOL_INFO_REGRELATIVE )
	{
		struc_t*	pstFrame = get_frame( ea );
		set_member_name( pstFrame, iOffset + get_func( ea )->frsize, pSymInfo->Name );
		giLocalVariableCount++;
	}

	return TRUE;
}


/***************************************************************************************************
*
*	FUNCTION		EnumerateSymbolsProc
*
*	DESCRIPTION		This callback has, at the present time, two operating modes. When the enumeration
*					mode variable is set to ENUMERATE_GLOBALS, then this function attempts to 
*					name items within global scope. Symbols that are tagged as functions are named
*					and also have 'add_func' called on them. When the enumeration mode is set to
*					ENUMERATE_LOCALS, then we look for global function symbols and use each function
*					as the starting point for a locally scoped enumeration that identifies local
*					symbols. This is achieved through the use of the SymSetContext() API call, prior
*					to the secondary call to SymEnumSymbols().
*
*	INPUTS			pSymInfo		-		A filled in SYMBOL_INFO structure.
*	
*					ulSymbolSize	-		Size of the symbol we're processing.
*
*					pvNothing		-		Not used in this callback.
*
*	RESULT			This function should return 'FALSE' if, for any reason, we need to stop the
*					symbol enumeration. Otherwise return 'TRUE'.
*
*	NOTES			I had to separate global & local scope enumerations because the calls to 
*					'add_func()' often cause IDA to resume autoanalysis. I need to be sure that
*					all autoanalysis-driven frame construction is completed before setting the names
*					of local symbols, or we'd end up not setting names for all variables.
*
***************************************************************************************************/

static BOOL CALLBACK EnumerateSymbolsProc( PSYMBOL_INFO pSymInfo, ULONG ulSymbolSize, PVOID pvNothing )
{
	ea_t ea		= ( ea_t )( pSymInfo->Address + giDelta );
	char *name	= pSymInfo->Name;

	// We have different behaviour depending on whether we're enumerating globals or locals.
	switch ( geEnumerationMode )
	{
		case	ENUMERATE_GLOBALS:
		{
			if ( pSymInfo->Tag == SymTagFunction )
			{
				// Global functions get a name, and also have a function created for them..
				giGlobalFunctionCount += set_name( ea, pSymInfo->Name, SN_NOWARN );
				add_func( ea, BADADDR );
			}
			else
			{
				// Global variables get a name.. everything else is up to IDA or the user.
				giGlobalVariableCount += set_name( ea, pSymInfo->Name, SN_NOWARN );
			}
			break;
		}

		case	ENUMERATE_LOCALS:
		{
			// We only go looking for locals within the context of a function..
			if ( pSymInfo->Tag == SymTagFunction )
			{
				IMAGEHLP_STACK_FRAME	obStackFrame;
				memset( &obStackFrame, 0, sizeof( obStackFrame ) );
				obStackFrame.InstructionOffset = pSymInfo->Address;
				SymSetContext( ghProcess, &obStackFrame, 0 );
				SymEnumSymbols( ghProcess, 0, 0, EnumerateLocalSymbolsProc, (PVOID)ea );
			}
			break;
		};
	}

	// Show where we are.. better than having no feedback at all.
	showAddr(ea);

	return TRUE;
}


/***************************************************************************************************
*
*	FUNCTION		PluginError
*
*	DESCRIPTION		Displays a system error. Used when something horrible happens..
*
*	INPUTS			pcName		-	Pointer to user-supplied error text.
*
***************************************************************************************************/

static void PluginError( char* pcName )
{
	LPVOID pvMessage;

	if ( FormatMessage(	FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,
						GetLastError(),
						MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
						( LPTSTR )&pvMessage,
						0, NULL ) )
	{
		msg( "%s: %s\n", pcName, pvMessage );
		LocalFree( pvMessage );
	}
	else
	{
		msg( "%s: Unknown error, get error message fail\n" );
	}
}


/***************************************************************************************************
*
*	FUNCTION		PluginMain
*
*	DESCRIPTION		This function contains the core plugin code.
*
***************************************************************************************************/

void idaapi PluginMain( int )
{
	char*	pcInputName;
	DWORD64	dwSymbolBase = 0;

	// Reset our symbol resolution counts
	giGlobalFunctionCount	= 0;
	giGlobalVariableCount	= 0;
	giLocalVariableCount	= 0;

	// Get the load base of the current PE file
	netnode	penode("$ PE header");
	ea_t	dwLoadedBase = penode.altval( -2 );

	// Get the input file name and try to guess the PDB file locaton If failed, ask the user
	pcInputName = get_input_file_path();

	if ( !qfileexist( pcInputName ) )
	{
		pcInputName = askfile_c( false, pcInputName, "Please specify the input file" );
		if ( !pcInputName )
			goto done;
	}

	// Initialise 
	if ( !SymInitialize( ghProcess, 0, FALSE ) )
	{
		PluginError( "SymInitialize Failed" );
		goto done;
	}

	// If I could find out how to validate the version of DBGHELP.DLL without having to manually
	// call LoadLibary() and GetProcAddress(), I'd put it here.. but for now I don't bother.
	// < Add code here.. maybe >	

	// Load the symbols..
	if ( ! ( dwSymbolBase = SymLoadModule64( ghProcess, 0, pcInputName, NULL, dwLoadedBase, 0 ) ) )
	{
		PluginError( "SymLoadModule64 Failed" );
		goto cleanup;
	}

	// Obtain symbols module information
	{
		IMAGEHLP_MODULE64	obModule = { sizeof( obModule ) };
		SymGetModuleInfo64( ghProcess, dwSymbolBase, &obModule );
		
		if ( obModule.SymType == SymNone )
		{
			PluginError( "SymGetModuleInfo64 : No Symbols Found");
			goto unload;
		}
	}

	// Construct a delta that might be needed..
	giDelta = ( int )( dwLoadedBase - dwSymbolBase );

	// Enumerate symbols to construct functions
	geEnumerationMode = ENUMERATE_GLOBALS;
	if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, 0, EnumerateSymbolsProc, 0 ) )
	{
		PluginError( "SymEnumSymbols(Functions) Failed" );
		goto unload;
	}
	msg( "PDBPlus : %d global function symbols loaded\n", giGlobalFunctionCount );
	msg( "PDBPlus : %d global variable symbols loaded\n", giGlobalVariableCount );

	// Wait for IDA to complete any pending autoanalysis before dealing with locals
	msg( "PDBPlus : Waiting for function analysis to complete\n" );
	autoWait();
	msg( "PDBPlus : Starting local symbol analysis\n" );

	// Enumerate symbols to construct local variables
	geEnumerationMode = ENUMERATE_LOCALS;
	if ( !SymEnumSymbols( ghProcess, ( DWORD )dwSymbolBase, 0, EnumerateSymbolsProc, 0 ) )
	{
		PluginError( "SymEnumSymbols(Variables) Failed" );
		goto unload;
	}
	msg( "PDBPlus : %d local variable symbols loaded\n", giLocalVariableCount );

unload:
	if ( !SymUnloadModule64( ghProcess, dwSymbolBase ) )
		PluginError( "SymUnloadModule64 Failed" );

cleanup:
	if ( !SymCleanup( ghProcess ) )
		PluginError( "SymCleanup Failed" );

done:
	return;
}


// -------------------------------------------------------------------------------------------------



/***************************************************************************************************
*
*	FUNCTION		PluginStartup
*
*	DESCRIPTION		IDA will call this function only once. If this function returns PLUGIN_SKIP, 
*					IDA will never load it again. If it returns PLUGIN_OK, IDA will unload the plugin
*					but remember that the plugin agreed to work with the database. The plugin will
*					be loaded again if the user invokes it by pressing the hotkey or selecting it
*					from the menu. After the second load, the plugin will stay in memory.
*
*	NOTES			
*
***************************************************************************************************/

int idaapi	PluginStartup(void)
{
	// We only support running our plugin when a PE file is loaded..
	if ( inf.filetype != f_PE )
		return PLUGIN_SKIP;

	return PLUGIN_OK;
}


/***************************************************************************************************
*
*	FUNCTION		PluginShutdown
*
*	DESCRIPTION		IDA will call this function when the user asks to exit. This function is *not*
*					called in the case of emergency exits.
*
*	NOTES			Normally this function is empty..
*
***************************************************************************************************/

void idaapi PluginShutdown(void)
{
}


/***************************************************************************************************
*
*	Strings required for IDA Pro's PLUGIN descriptor block			
*
***************************************************************************************************/

char	gacPluginComment[]		=	"Load global & local debug information from a PDB file";
char	gacPluginShortName[]	=	"PDB Plus";
char	gacPluginHelpText[]		=	"PDB Plus\n"
									"\n"
									"This module allows you to load debug information about function\n"
									"names, local variables, data names, and structures from a PDB file.\n"
									"\n"
									"The PDB file should be in the same directory as the input file\n"
									"\n"
									"Note that a recent version of DBGHELP.DLL (5.0+) is required\n";


/***************************************************************************************************
*
*	This 'PLUGIN' data block is how IDA Pro interfaces with this plugin. 
*
***************************************************************************************************/

extern "C" plugin_t PLUGIN =
{
	IDP_INTERFACE_VERSION,		// Version
	PLUGIN_UNL,					// Plugin Flags: unload the plugin immediately after run
	PluginStartup,				// Plugin Startup
	PluginShutdown,				// Plugin Shutdown.. this can be NULL
	PluginMain,					// Plugin Main

	gacPluginComment,			// Comment about the plugin
	gacPluginHelpText,			// Multiline help about the plugin

	gacPluginShortName,			// Preferred short name of the plugin
	"",							// Preferred hotkey to run the plugin (no hotkey specified here)
};


